from pydantic import BaseModel
from typing import get_type_hints, Optional, Union


def type_to_str(tp):
    """Convert a type to a string representation."""
    # Handle Optional types (Union[..., None])
    origin = getattr(tp, "__origin__", None)
    args = getattr(tp, "__args__", ())

    if origin is Union and type(None) in args:
        non_none = [a for a in args if a is not type(None)]

        if len(non_none) == 1:
            return f"Optional[{type_to_str(non_none[0])}]"

    if hasattr(tp, "__name__"):
        return tp.__name__

    if hasattr(tp, "_name"):  # for generic types like List, Dict
        args_str = ", ".join(type_to_str(a) for a in args)
        return f"{tp._name}[{args_str}]"

    return str(tp).replace("typing.", "")


def generate_pydantic_code(model_cls: type[BaseModel]) -> str:
    """
    Generate Python source code for a Pydantic BaseModel subclass.
    """
    lines = []
    visited = set()

    def generate(cls, lines: list[str], visited: set):
        class_name = model_cls.__name__

        if cls.__name__ in visited:
            return

        visited.add(cls.__name__)

        # Generate the class definition
        lines.append(f"class {class_name}(BaseModel):")

        subtypes = []

        # Get type hints (annotations) for the model
        hints = get_type_hints(model_cls)

        # For each field, get name, type, and default value if any
        for field_name, field in model_cls.model_fields.items():
            field_type = hints.get(field_name, "Any")
            type_str = type_to_str(field_type)

            # Determine default value
            if not field.is_required:
                default_val = repr(field.default)
                line = f"    {field_name}: {type_str} = {default_val}"
            else:
                line = f"    {field_name}: {type_str}"

            lines.append(line)

            # if its a BaseModel, add to visited
            if issubclass(field_type, BaseModel):
                subtypes.append(field_type)

        if subtypes:
            lines.append("")

        for cls in subtypes:
            generate(cls, lines, visited)

    generate(model_cls, lines, visited)

    return "\n".join(lines)



from pydantic import BaseModel, AnyUrl, Field
import yaml
import pathlib


class ModelConfig(BaseModel):
    base_url: str = Field(default="", description="The API base URL (in case you're not using OpenAI)")
    api_key: str = Field(default="", description="The API key to authenticate with the LLM provider")
    model: str = Field(default="", description="The concrete LLM model to use")


class LovelaiceConfig(BaseModel):
    chat_model: ModelConfig = Field(default_factory=ModelConfig, description="Configuration for the default chat model.")
    audio_model: ModelConfig = Field(default_factory=ModelConfig, description="Configuration for the audio model.")
    code_model: ModelConfig = Field(default_factory=ModelConfig, description="Configuration for the code model.")
    max_tokens: int = Field(2048, description="Max number of tokens to generate in a single prompt")
    min_words: int = Field(0, description="For completion only, min number of words to generate")

    @classmethod
    def load(cls, root_path: pathlib.Path = pathlib.Path(".")):
        root_path = root_path.absolute()

        if not root_path.is_dir():
            raise ValueError('The root path must a directory.')

        matches: list[pathlib.Path] = []

        while True:
            if (root_path / ".lovelaice.yml").exists():
                matches.append(root_path / ".lovelaice.yml")

            if root_path.parent != root_path:
                root_path = root_path.parent
            else:
                break

        config = {}

        while matches:
            path = matches.pop()

            with path.open() as fp:
                config.update(yaml.safe_load(fp))

        return LovelaiceConfig(**config)


    def save(self, root_path: pathlib.Path = pathlib.Path(".")):
        root_path = root_path.absolute()

        if not root_path.is_dir():
            raise ValueError('The root path must a directory.')

        with open(root_path / ".lovelaice.yml", "w") as fp:
            yaml.dump(self.model_dump(mode="json"), fp)



import abc
from typing import Annotated, Any, Union
import yaml

from pydantic import BaseModel, Discriminator, Field, RootModel, Tag, model_validator

from .agent import ChatAgent
from .skills import Skill, chat
from .tools import Tool
from .llm import LLM, Message
from .context import Context


PREDEFINED_SKILLS = {
    "chat": chat
}


class ToolConfig(BaseModel):
    name: str
    description: str


class SkillStep(BaseModel):
    @abc.abstractmethod
    def compile(self):
        pass


class DecideStep(SkillStep):
    decide: str | None
    yes: "StepList"
    no: "StepList"

    def compile(self):
        true_branch = self.yes.compile()
        false_branch = self.no.compile()

        async def decide_step(ctx: Context) -> Message:
            instructions = []

            if self.decide:
                instructions.append(Message.system(self.decide))

            decision = await ctx.decide(*instructions)

            if decision:
                return await true_branch(ctx)
            else:
                return await false_branch(ctx)

        return decide_step


class ChooseStep(SkillStep):
    choose: str | None = None
    choices: dict[str, "StepList"]

    @model_validator(mode="before")
    def validate(cls, data):
        if isinstance(data, dict):
            choose = data.pop("choose")
            return dict(choose=choose, choices=data)

        return data

    def compile(self):
        compiled_choices = {k: v.compile() for k, v in self.choices.items()}

        async def choose_step(ctx: Context) -> Message:
            instructions = []

            if self.choose:
                instructions.append(Message.system(self.choose))

            choice = await ctx.choose(list(compiled_choices.keys()), *instructions)

            return await compiled_choices[choice](ctx)

        return choose_step


class WhileStep(SkillStep):
    condition: str
    steps: "StepList"

    @model_validator(mode="before")
    def validate(cls, data):
        if isinstance(data, dict):
            return dict(condition=data['while'], steps=data['do'])

        return data

    def compile(self):
        compiled_steps = self.steps.compile()

        async def while_step(ctx: Context) -> Message:
            m: Message = await compiled_steps(ctx)
            ctx.add(m)

            while await ctx.decide(self.condition):
                m = await compiled_steps(ctx)
                ctx.add(m)

            return m

        return while_step


class UntilStep(SkillStep):
    condition: str
    steps: "StepList"

    @model_validator(mode="before")
    def validate(cls, data):
        if isinstance(data, dict):
            return dict(condition=data['until'], steps=data['do'])

        return data

    def compile(self):
        compiled_steps = self.steps.compile()

        async def until_step(ctx: Context) -> Message:
            m: Message = await compiled_steps(ctx)
            ctx.add(m)

            while not await ctx.decide(self.condition):
                m = await compiled_steps(ctx)
                ctx.add(m)

            return m

        return until_step


class ReplyStep(SkillStep):
    reply: str | None

    def compile(self):
        async def reply_step(ctx: Context) -> Message:
            instructions = []

            if self.reply:
                instructions.append(Message.system(self.reply))

            return await ctx.reply(*instructions)

        return reply_step


def get_skill_step_discriminator_value(v: Any) -> str:
    if isinstance(v, SkillStep):
        return v.__class__.__name__
    elif isinstance(v, dict):
        if "decide" in v:
            return "DecideStep"
        elif "choose" in v:
            return "ChooseStep"
        elif "reply" in v:
            return "ReplyStep"
        elif "while" in v:
            return "WhileStep"
        elif "until" in v:
            return "UntilStep"

    raise ValueError(f"Invalid SkillStep: {v}")


class StepList(
    RootModel[
        list[
            Annotated[
                Union[
                    Annotated[DecideStep, Tag("DecideStep")],
                    Annotated[ChooseStep, Tag("ChooseStep")],
                    Annotated[ReplyStep, Tag("ReplyStep")],
                    Annotated[WhileStep, Tag("WhileStep")],
                    Annotated[UntilStep, Tag("UntilStep")],
                ],
                Discriminator(get_skill_step_discriminator_value),
            ]
        ]
    ]
):
    pass

    def compile(self):
        steps = [s.compile() for s in self.root]

        async def step_list(ctx: Context) -> Message:
            m: Message | None = None
            messages = ctx.messages

            for step in steps:
                m = await step(ctx)
                messages.append(m)

            if m is None:
                raise ValueError("No message returned from steps")

            return m

        return step_list


class SkillConfig(BaseModel):
    name: str
    description: str
    steps: StepList

    def compile(self) -> Skill:
        return DeclarativeSkill(self)


class DeclarativeSkill(Skill):
    def __init__(self, config: SkillConfig):
        super().__init__(config.name, config.description)
        self.steps = config.steps.compile()

    async def _execute(self, ctx):
        return await self.steps(ctx)


class AgentConfig(BaseModel):
    name: str
    description: str

    tools: list[ToolConfig] = Field(default_factory=list)
    skills: list[SkillConfig | str]

    def compile(self, llm: LLM) -> ChatAgent:
        agent = ChatAgent(name=self.name, description=self.description, llm=llm)

        for s in self.skills:
            if isinstance(s, str):
                skill = PREDEFINED_SKILLS[s]
            else:
                skill = s.compile()

            agent.skill(skill)

        return agent


def _fix_dumb_yes_no(item):
    def f(x):
        if x is True:
            return "yes"
        if x is False:
            return "no"
        return x

    if isinstance(item, list):
        return [_fix_dumb_yes_no(x) for x in item]
    if isinstance(item, dict):
        return {f(k): _fix_dumb_yes_no(v) for k, v in item.items()}

    return item


def parse(path) -> AgentConfig:
    with open(path) as fp:
        config = yaml.safe_load(fp)
        config = _fix_dumb_yes_no(config)
        return AgentConfig(**config) # type: ignore



import asyncio
import collections
import random
from typing import Literal

from argo.agent import AgentBase
from argo.llm import LLM, Message
from pydantic import BaseModel
from argo.crew import Crew, MemoryBoard

import os
import dotenv

dotenv.load_dotenv()


class EmailStartup(BaseModel):
    username: str


class EmailItem(BaseModel):
    sender: str
    body: str


class EmailSummary(BaseModel):
    summary: str
    sentiment: str
    category: Literal['personal', 'work', 'promotional', 'subscription', 'spam']


class EmailAggregation(BaseModel):
    summary: str
    category: str


class EmailSummaries(BaseModel):
    summaries: list[EmailSummary]


llm = LLM(
    model=os.getenv("MODEL"),
    api_key=os.getenv("API_KEY"),
    base_url=os.getenv("BASE_URL"),
    verbose=True,
)


class Generator(AgentBase[EmailStartup, EmailItem]):
    """
    Generates random emails.
    """

    async def process(self, input: EmailStartup):
        while True:
            category = random.choice(['personal', 'work', 'promotional', 'subscription', 'spam'])
            sender = random.choice(["Alice", "Bob", "Charlie", "Dave"])

            yield await llm.create(
                EmailItem,
                messages=[
                    Message.system(
                        f"""
Create a random {category} email example directed to {input.username} from {sender}.

Reply only in JSON format with keys sender and body.
"""
                    )
                ],
                temperature=2,
            )
            await asyncio.sleep(random.randint(1, 5))


class Summarizer(AgentBase[EmailItem, EmailSummary]):
    """
    Summarizes incoming emails.
    """

    async def process(self, input: EmailItem):
        yield await llm.create(
            EmailSummary,
            messages=[
                Message.system(
                    """
Summarize the following email.
Reply only in JSON format,
with the following keys: summary, sentiment, category.
"""
                ),
                Message.user(input.model_dump_json()),
            ],
        )


class Aggregator(AgentBase[EmailSummary, EmailAggregation]):
    """
    Aggregates email summaries into categories.
    """

    def __init__(self, min_count:int):
        self.min_count = min_count
        self.summaries = collections.defaultdict(list)

    async def process(self, input: EmailSummary):
        self.summaries[input.category].append(input)

        if len(self.summaries[input.category]) >= self.min_count:
            yield await llm.create(
                EmailAggregation,
                messages=[
                    Message.system(
                        """
Create a summary of the following emails.
Reply only in JSON format,
with the following keys: category, summary.
"""
                    ),
                    Message.user(
                        EmailSummaries(summaries=self.summaries[input.category])
                    ),
                ],
            )
            self.summaries[input.category] = []


crew = Crew(
    board=MemoryBoard(),
    agents=[Generator(), Summarizer(), Aggregator(min_count=5)],
    seed=[EmailStartup(username="John")],
)

crew.run()



from pydantic import BaseModel
from argo import ChatAgent, LLM, Message, Context
import dotenv
import os
import wikipedia

from argo.cli import loop


dotenv.load_dotenv()


def callback(chunk: str):
    print(chunk, end="")


agent = ChatAgent(
    name="Trivial",
    description="A helpful assistant that can search Wikipedia for answering factual questions.",
    llm=LLM(model=os.getenv("MODEL"), callback=callback, verbose=True),
)


@agent.skill
async def chat(ctx: Context):
    """Casual chat with the user.

    Use this only for greetings, basic chat,
    and questions regarding your own capabilities.
    """
    yield await ctx.reply()


class Reasoning(BaseModel):
    observation: str
    thought: str
    query: str | None = None
    final: bool


class Summary(BaseModel):
    summary: str
    relevant: bool


@agent.skill
async def question_answering(ctx: Context):
    """Answer questions about the world.

    Use this skill when the user asks any questions
    that might require external knowledge.
    """

    ctx.add(
        """
        You have access to Wikipedia's search engine to answer the user quesstion.
        Do not assume you know anything about the question, always double check
        against Wikipedia.
        """
    )

    for i in range(5):
        reasoning = await ctx.create(
            """
            Breakdown the user request.
            First provide an observation of the
            current state of the task and the knowledge you already have.
            Then, provide a thought on the next step to take.
            Finally, provide a short, concise query for Wikipedia, if necessary.
            Otherwise, if the existing information is enough to answer, set final=True.
            """,
            model=Reasoning,
        )

        ctx.add(reasoning)

        if reasoning.final:
            yield await ctx.reply()

        results = await ctx.invoke(search, errors="handle")

        ctx.add(results)

    yield await ctx.reply("Reply with the best available information in the context.")


@agent.tool
async def search(query: str, llm: LLM) -> list[Summary]:
    """Search Wikipedia for information."""
    candidates = wikipedia.search(query, results=10)

    results = []

    for title in candidates:
        page = wikipedia.page(title)
        summary = await llm.create(
            messages=[
                Message.system(
                f"""
                Given the following query, summarize the information
                in the text to answer the query.

                Query: {query}

                Reply with a JSON object with the following fields:
                - summary: A short summary of the information relevant to the query.
                - relevant: A boolean indicating if the information is relevant to the query.
                """
                ),
                Message.user(page.summary),
            ],
            model=Summary,
        )

        if summary.relevant:
            results.append(summary)

    return results


loop(agent)



import abc

from .llm import Message
from typing import AsyncIterator



class Skill:
    def __init__(self, name: str, description: str):
        self._name = name
        self._description = description

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @abc.abstractmethod
    async def execute(self, ctx) -> AsyncIterator[Message]:
        pass


class MethodSkill(Skill):
    def __init__(self, name: str, description: str, target):
        super().__init__(name, description)
        self._target = target

    async def execute(self, ctx): # type: ignore
        async for m in self._target(ctx):
            yield m


async def chat(ctx: "Context"):
    """
    Casual chat with the user.
    """
    yield await ctx.reply()



from argo import ChatAgent, LLM
from argo.cli import loop
from argo.skills import chat
import dotenv
import os


dotenv.load_dotenv()


def callback(chunk:str):
    print(chunk, end="")


agent = ChatAgent(
    name="Agent",
    description="A helpful assistant.",
    llm=LLM(model=os.getenv("MODEL"), callback=callback, verbose=False),
    skills=[chat],

)


loop(agent)



import subprocess
import importlib
import requests
from pathlib import Path


class Tool:
    skip_use = False

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def description(self) -> str:
        return self.__class__.__doc__

    def describe(self) -> str:
        return f"- {self.name}: {self.description}."

    def prompt(self, query) -> str:
        return query

    def use(self, query, response):
        pass

    def conclude(self, query, output):
        pass


class Chat(Tool):
    """
    When the user engages in general-purpose or casual conversation.
    """

    def __init__(self) -> None:
        self.skip_use = True


class Bash(Tool):
    """
    When the user requests some action in the filesystem or terminal,
    including git commands, or installing new applications or packages.
    """

    def prompt(self, query) -> str:
        return f"""
Given the following user query, generate a single bash command line
that performs the indicated functionality.

Reply only with the corresponding bash line.
Do not add any explanation.

Query: {query}
Command:
"""

    def use(self, query, response):
        response = response.strip("`")

        if response.startswith("bash"):
            response = response[4:]

        response = response.split("`")[0]
        response = [s.strip() for s in response.split("\n")]
        response = [s for s in response if s]

        response = ";".join(s for s in response)

        yield "Running the following code:\n"
        yield "$ "
        yield response
        yes = input("\n[y]es / [N]o ")

        if yes != "y":
            yield "(!) Operation cancelled by your request.\n"
            return

        yield "\n"

        p = subprocess.run(response, shell=True, stdout=subprocess.PIPE)
        yield p.stdout.decode("utf8")

    def conclude(self, query, output):
        return f"""
The user issued the following query:

Query: {query}

Given that query, you ran the following command which
produced the given output:

---
{output}
---

If the user query is a question, answer it as succintly
as possible given the output.

If the user query was a request to do something,
explain briefly the result of the operation.
"""


class Interpreter(Tool):
    """
    When the user asks a mathematical question that can
    be solved with a simple Python function.
    """

    def prompt(self, query) -> str:
        return f"""
Given the following user query,
generate a single Python function named `solve`
and the necessary import statements
to perform the indicated functionality.

If you need secondary functions, name them starting with `_`.

Enclose the code in ```python and ```

Reply only with the corresponding Python code.
Do not add any explanation.
Do not execute the function.
Do not add any print statements.

Query: {query}
Function definition:
"""

    def use(self, query, response):
        code = []
        imports = []
        inside = False

        for line in response.split("\n"):
            if line.startswith("```python"):
                inside = True
            elif line.startswith("```"):
                inside = False
            elif inside:
                if line.startswith("import"):
                    imports.append(line.split()[1])
                else:
                    code.append(line)

        code.append("\nresult = solve()")
        code = "\n".join(code).strip()

        yield "Will run the following code:\n\n"
        yield code
        yes = input("\n\n[y]es / [N]o ")

        if yes != "y":
            yield "(!) Operation cancelled by your request.\n"
            return

        globals = {module: importlib.import_module(module) for module in imports}
        locals = {}
        exec(code, globals, locals)
        result = locals["result"]

        yield f"\nResult: {result}"


class Codegen(Tool):
    """
    When the user makes a general question about programming or
    explicitly asks to generate code in a given programming language.
    """

    def __init__(self) -> None:
        self.skip_use = True

    def prompt(self, query) -> str:
        return f"""
Answer the following user query about programming with
a general explanation in broad terms, followed by
one or more examples of code, as necessary.

Enclose all code examples in ``` with the corresponding
programming language identifier.

Query: {query}
"""

class Weather(Tool):
    """
    When the user asks for the weather.
    """

    def prompt(self, query) -> str:
        yaml_content = importlib.resources.files('lovelaice').joinpath('api.open-meteo.yml').read_text()
        return f"""
You're a helpful weather assistant. I want to retrieve the weather for a user query using this OpenAPI schema: 
```yaml
{yaml_content}
```

If the user requests the weather for:
- today, focus in the `current` parameter.
- a specific day, focus in the `hourly` parameter.
- a future day, focus in the `daily` parameter and use `forecast_days` parameter.
- a past day, focus in the `hourly` parameter and use `past_days` parameter.
- a range of days, focus in the `start_date` and `end_date` parameters.

If the user requests the weather in general, focus in temperature, wind, humidity, precipitation, cloud cover, and weather codes if available.

Generate the URL to retrieve the weather for the user query using the right parameters.
Do not add any other text than the URL.

Query: {query}
URL: 
"""

    def use(self, query, response):
        endpoint = response.split()[1]
        yield f"Retrieving weather from {endpoint}\n\n"

        response = requests.get(endpoint)
        yield response.text.strip() + "\n"
    
    def conclude(self, query, output):
        return f"""
The user issued the following query:

Query: {query}

The weather for the user query is: {output}. Write a short summary of the weather. If the temperature appears in the response, show itin Celsius and Fahrenheit.
"""


import asyncio
import sys

from pathlib import Path
import dotenv
import rich
from typer import Argument, Option, Typer, Exit
from .agent import ChatAgent
from .llm import LLM, Message
from .declarative import parse


app = Typer(name="argo", help="Argo CLI", no_args_is_help=True)


def loop(agent: ChatAgent, print_response: bool = False):
    """Runs a CLI agent loop with integrated
    conversation history management.

    This method creates an async context internally,
    and handles storing and retrieving conversation history.
    It also handles keyboard interrupts and EOF errors.

    This method blocks the terminal waiting for user input,
    and loops until EOF (Ctrl+D) is pressed.
    """
    rich.print(f"[bold green]{agent.name}[/bold green]: {agent.description}\n")

    if agent.llm.verbose:
        rich.print(f"[yellow]Running in verbose mode.[/yellow]")

    rich.print(f"[yellow]Press Ctrl+D to exit at any time.\n[/yellow]")

    async def run():
        while True:
            try:
                user_input = input(">>> ")
                m = Message.user(user_input)

                async for m in agent.perform(m):
                    pass

                if print_response:
                    rich.print(m)

                print("\n")
            except (EOFError, KeyboardInterrupt):
                break

    asyncio.run(run())


@app.command()
def run(
    path: Path = Argument(help="A YAML file with an agent definition."),
    api_key: str = Option(
        None, "--api-key", "-k", help="API key for the LLM.", envvar="API_KEY"
    ),
    base_url: str = Option(
        None, "--base-url", "-u", help="Base URL for the LLM.", envvar="BASE_URL"
    ),
    model: str = Option(
        ..., "--model", "-m", help="Model to use for the LLM.", envvar="MODEL"
    ),
):
    """
    Run an agent defined in a YAML file with a basic CLI loop.
    """
    def callback(chunk: str):
        print(chunk, end="")

    llm = LLM(model=model, api_key=api_key, base_url=base_url, callback=callback)

    config = parse(path)
    agent = config.compile(llm)
    loop(agent)


@app.command()
def serve(
    path: Path = Argument(help="A YAML file with an agent definition."),
    api_key: str = Option(
        None, "--api-key", "-k", help="API key for the LLM.", envvar="API_KEY"
    ),
    base_url: str = Option(
        None, "--base-url", "-u", help="Base URL for the LLM.", envvar="BASE_URL"
    ),
    model: str = Option(
        ..., "--model", "-m", help="Model to use for the LLM.", envvar="MODEL"
    ),
    host: str = Option("127.0.0.1", "--host", "-h", help="Host IP to bind to."),
    port: int = Option(8000, "--port", "-p", help="Port to bind to."),
):
    """
    Start a FastAPI server to run an agent in API-mode.
    """
    try:
        from .server import serve as serve_loop
    except ImportError:
        print("Please install argo[server] to use this command.")
        raise Exit(1)

    llm = LLM(model=model, api_key=api_key, base_url=base_url)

    config = parse(path)
    agent = config.compile(llm)
    serve_loop(agent, host=host, port=port)


def main():
    app()


if __name__ == "__main__":
    main()



import functools
import inspect
from typing import Any, Callable, Coroutine, Literal
import rich
import openai
from pydantic import BaseModel
import os

from openai.types.chat import ChatCompletionMessageParam


class Message(BaseModel):
    role: Literal["user", "system", "assistant", "tool"]
    content: Any

    @classmethod
    def system(cls, content: Any) -> "Message":
        return cls(role="system", content=content)

    @classmethod
    def user(cls, content: Any) -> "Message":
        return cls(role="user", content=content)

    @classmethod
    def assistant(cls, content: Any) -> "Message":
        return cls(role="assistant", content=content)

    @classmethod
    def tool(cls, content: Any) -> "Message":
        return cls(role="tool", content=content)

    def dump(self) -> ChatCompletionMessageParam:
        return dict(  # type: ignore
            role=self.role,
            content=(
                self.content.model_dump_json()
                if isinstance(self.content, BaseModel)
                else str(self.content)
            ),
        )

    def unpack[T: BaseModel](self, t: type[T]) -> T:
        if isinstance(self.content, t):
            return self.content

        if isinstance(self.content, str):
            return t.model_validate_json(self.content)

        raise TypeError(f"Cannot unpack {self.content} into {t}")


class LLM:
    def __init__(
        self,
        model: str,
        callback: Callable[[str], None] | None = None,
        verbose: bool = False,
        base_url: str | None = None,
        api_key: str | None = None,
    ):
        self.model = model
        self.verbose = verbose

        if base_url is None:
            base_url = os.getenv("BASE_URL")
        if api_key is None:
            api_key = os.getenv("API_KEY")

        self.client = openai.AsyncOpenAI(base_url=base_url, api_key=api_key)
        self.callback = callback

    async def chat(self, messages: list[Message], **kwargs) -> Message:
        result = []

        async for chunk in await self.client.chat.completions.create(
            model=self.model,
            messages=[message.dump() for message in messages],
            stream=True,
            **kwargs,
        ):
            content = chunk.choices[0].delta.content

            if content is None:
                continue

            if self.callback:
                if inspect.iscoroutinefunction(self.callback):
                    await self.callback(content)
                else:
                    self.callback(content)

            result.append(content)

        return Message.assistant("".join(result))

    async def create[T: BaseModel](
        self, model: type[T], messages: list[Message], **kwargs
    ) -> T:
        response = await self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[message.dump() for message in messages],
            response_format=model,
            **kwargs,
        )

        result = response.choices[0].message.parsed

        if self.verbose:
            rich.print(result)

        if result is None:
            raise ValueError("Failed to parse the response.")

        return result

    def wrap(self, target):
        llm_param = None
        parameters = inspect.signature(target).parameters

        for name, param in parameters.items():
            if issubclass(param.annotation, LLM):
                llm_param = name
                break

        if llm_param is None:
            raise TypeError("LLM parameter not found.")

        @functools.wraps(target)
        async def wrapper(*args, **kwargs):
            kwargs[llm_param] = self
            return await target(*args, **kwargs)

        # Remove the LLM param from wrapper so future
        # introspection doesn't see this parameter
        wrapper.__annotations__.pop(llm_param)

        return wrapper



from argo import ChatAgent, LLM, Message, Context
import dotenv
import os
import rich
from argo.cli import loop


dotenv.load_dotenv()


def callback(chunk: str):
    print(chunk, end="")


agent = ChatAgent(
    name="Coder",
    description="A helpful assistant that can write and run Python code.",
    llm=LLM(model=os.getenv("MODEL"), callback=callback),
)


@agent.skill
async def code(ctx: Context):
    """Use Python to compute math operations.

    Use this skill when you need to compute some math operations.
    """
    result = await ctx.invoke(interpreter)
    yield await ctx.reply(f"Result={result}.\n\nReply to the user.")


@agent.tool
async def interpreter(code: str):
    """Run Python code and returns a final value.

    The code should compute and store the result
    in a variable called `result`.

    You can import built-in modules, define intermediate
    functions, classes, and variables.

    The final statement should be assigning `result`.

    Do not print anything to the console or use
    non-standard libraries.
    """
    rich.print(f"```python\n{code}\n```")

    env = {}
    exec(code, env, env)
    return env.get("result")


loop(agent)



import abc
import asyncio
import threading

from .agent import Agentic
from .llm import Message


class MessageBoard(abc.ABC):
    """
    Represents a message board.

    Provides methods to listen to messages and post messages.
    Messages are parameterized by type.
    """

    @abc.abstractmethod
    async def get[T](self, message_type: type[T]) -> T:
        pass

    @abc.abstractmethod
    async def post[T](self, message: T):
        pass


class MemoryBoard(MessageBoard):
    """
    A simple in-memory message board that uses asyncio queues.
    """

    def __init__(self):
        self.queues: dict[type, asyncio.Queue] = {}

    async def get[T](self, message_type: type[T]) -> T:
        if message_type not in self.queues:
            self.queues[message_type] = asyncio.Queue()

        return await self.queues[message_type].get()

    async def post[T](self, message: T):
        if type(message) not in self.queues:
            self.queues[type(message)] = asyncio.Queue()

        await self.queues[type(message)].put(message)


class Crew:
    """
    Represents a crew of agents.

    Each agent will be run asynchronously and will listen to messages of a specific type
    (as declared by the `Agentic` protocol).

    The crew will post messages to the message board that will be automatically
    picked up by the right agent if such agent exists.
    """

    def __init__(self, board: MessageBoard, agents: list[Agentic], seed:list=None):
        self.agents = agents
        self.board = board
        self.seed = seed or []


    async def loop(self):
        """
        Starts the crew loop.

        This method blocks in the current thread and must be called
        from an async context.

        Use `start()` if you want to run the loop in a background thread.
        """
        for item in self.seed:
            await self.board.post(item)

        try:
            await asyncio.gather(*[self._loop_agent(agent) for agent in self.agents])
        except KeyboardInterrupt:
            pass

    def start(self):
        """
        Runs `loop()` as a background async task and returns the task.
        """
        return asyncio.create_task(self.loop())

    async def _loop_agent(self, agent: Agentic):
        """
        Starts the loop for a single agent.
        """
        in_t, out_t = agent.types

        while True:
            m = await self.board.get(in_t)

            async for m in agent.perform(Message.system(m)):
                await self.board.post(m.content)

    def run(self):
        """
        Blocks the current thread until the crew loop is stopped.
        """
        asyncio.run(self.loop())



import os
import dotenv
import asyncio
import argparse
from pydantic import BaseModel
from rich.prompt import Prompt, Confirm
from rich import print
from .core import Lovelaice
from argo.cli import loop
from argo import LLM
from .tools import Bash, Chat, Codegen, Interpreter, Weather
from .config import LovelaiceConfig


def run():
    dotenv.load_dotenv()
    config = LovelaiceConfig.load()

    parser = argparse.ArgumentParser("lovelaice", usage="lovelaice [options] query ...")
    parser.add_argument("--version", action="store_true", help="Print version and exit")
    parser.add_argument(
        "--config", action="store_true", help="Run configuration and exit"
    )
    parser.add_argument(
        "-c",
        "--complete",
        action="store_true",
        help="Instead of full chat interaction, simply run completion on the input prompt.",
        default=False,
    )
    parser.add_argument(
        "-cf", "--complete-files",
        action="store",
        nargs="*",
        help="Similar to completion mode, but instead CLI, it will read these files and replace all instances of `+++` with a completion, using the previous content as prompt.",
        metavar="FILE",
        default=None,
    )
    parser.add_argument(
        "-w", "--watch",
        action="store_true",
        help="Used only with --complete-files, keep watching for file changes until stopped with Ctrl+C.",
        default=None,
    )
    parser.add_argument(
        "-f", "--file", action="store", help="Add a file to the context"
    )
    parser.add_argument(
        "-a", "--audio", action="store", help="Add an audio file to the context"
    )
    parser.add_argument(
        "-j", "--json", action="store_true", help="Force the reply to JSON mode"
    )

    parser.add_argument("--api", action="store_true", help="Run an HTTP server instead of CLI.", default=False)
    parser.add_argument("--host", action="store", help="Host to bind the API.", default="127.0.0.1")
    parser.add_argument("--port", action="store", type=int, help="Port to bind the API.", default=8000)
    parser.add_argument("--debug", action="store_true", help="Runs in debug mode, e.g. more verbose.", default=False)
    parser.add_argument("query", nargs="*", default=None)

    args = parser.parse_args()

    if args.config:
        configure(config)
        return

    def callback(chunk:str):
        print(chunk, end="")

    llm = LLM(model=config.chat_model.model, api_key=config.chat_model.api_key, base_url=config.chat_model.base_url, callback=callback)

    if args.complete:
        asyncio.run(complete(args, config, llm))
        return

    if args.complete_files:
        asyncio.run(complete_files(args, config, llm))
        return

    agent = Lovelaice(
        llm
    )

    if args.query:
        asyncio.run(run_once(args, config, agent))
    else:
        loop(agent)


def _build_config(model: type[BaseModel], old_config, indent=0):
    new_config = {}

    for field, info in model.model_fields.items():
        if isinstance(info.annotation, type) and issubclass(info.annotation, BaseModel):
            print(f"[purple]{indent * "  "}{field}[/purple]: {info.description}\n")
            value = _build_config(info.annotation, old_config.get(field, {}), indent+1)
        else:
            print(f"[yellow]{indent * "  "}{info.description}[/yellow]")
            value = Prompt.ask(indent * "  " + field, default=str(old_config[field]))
            print()

        if value is not None:
            new_config[field] = value

    return new_config


def configure(config: LovelaiceConfig):
    old_config = config.model_dump(mode="json")
    new_config = _build_config(LovelaiceConfig, old_config)

    new_config = LovelaiceConfig(**new_config)
    print(new_config.model_dump(mode="json"))

    if Confirm.ask("Are you happy with this configuration?"):
        new_config.save()


async def complete(args, config: LovelaiceConfig, llm: LLM):
    prompt = " ".join(args.query)

    print(prompt, end="", flush=True)

    while True:
        generated = False

        async for chunk in llm.complete_stream(prompt, max_tokens=config.max_tokens):
            prompt += chunk
            print(chunk, end="", flush=True)

            if chunk:
                generated = True

        if not generated or len(prompt.split()) > config.min_words:
            break

    print()


async def _detect_file_changes(file, interval=1):
    last_modified = os.path.getmtime(file)

    while True:
        current_modified = os.path.getmtime(file)

        if current_modified != last_modified:
            return

        await asyncio.sleep(interval)


async def complete_files(args, config, llm: LLM):
    for file in args.complete_files:
        await _complete_file(file, args, llm)


async def _complete_file(file, args, config: LovelaiceConfig, llm: LLM):
    while True:
        prompt = []
        complete = False

        with open(file) as fp:
            for line in fp:
                if line.strip().endswith("+++"):
                    if line.strip() != "+++":
                        prompt.append(line.replace("+++", ""))
                    complete = True
                    break
                else:
                    prompt.append(line)

        prompt = "\n".join(prompt)

        if prompt and complete:
            print(f"Running completion on {file}...")
            response = await llm.complete(prompt, max_tokens=config.max_tokens)
            print(f"Done with completion on {file}.")

            lines = open(file).readlines()

            with open(file, "w") as fp:
                for line in lines:
                    if "+++" in line:
                        line = line.replace("+++", response)

                    fp.write(line)
        else:
            print(f"Nothing to do in {file}.")

        if args.watch:
            print(f"Waiting for changes on {file}.")
            await _detect_file_changes(file)
        else:
            return


async def run_once(args, config: LovelaiceConfig, agent: Lovelaice):
    prompt = " ".join(args.query)
    transcription = ""

    if args.audio:
        with open(args.audio, 'rb') as fp:
            transcription = await agent.client.transcribe(fp)

    prompt = "\n\n".join([transcription, prompt])

    async for response in agent.query(prompt, max_tokens=config.max_tokens):
        print(response, end="", flush=True)

    print()



import inspect
import abc
from typing import AsyncIterator, Protocol

from .llm import LLM, Message
from .prompts import DEFAULT_SYSTEM_PROMPT
from .skills import Skill, MethodSkill
from .tools import Tool, MethodTool


class Agentic(Protocol):
    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def description(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def types(self) -> tuple[type, type]:
        pass

    @abc.abstractmethod
    def perform(self, input: Message) -> AsyncIterator[Message]:
        pass


class AgentBase[In, Out](Agentic):
    @property
    def name(self):
        return self.__class__.__name__

    @property
    def description(self):
        return self.__class__.__doc__ or ""

    @property
    def types(self):
        return self.__class__.__orig_bases__[0].__args__ # type: ignore

    async def perform(self, input: Message) -> AsyncIterator[Message]:
        in_t, _ = self.types
        data: In = input.unpack(in_t)

        async for m in self.process(data):
            yield Message.assistant(m)

    @abc.abstractmethod
    def process(self, input: In) -> AsyncIterator[Out]:
        pass


class ChatAgent(Agentic):
    def __init__(
        self,
        name: str,
        description: str,
        llm: LLM,
        *,
        system_prompt=DEFAULT_SYSTEM_PROMPT,
        persistent:bool=True,
        skills: list | None = None,
        tools: list | None = None
    ):
        self._name = name
        self._description = description
        self._llm = llm
        self._skills = []
        self._tools = []
        self._system_prompt = system_prompt.format(name=name, description=description)
        self._conversation = [Message.system(self._system_prompt)]
        self._persistent = persistent

        # initialize predefined skills and tools
        for skill in skills or []:
            self.skill(skill)

        for tool in tools or []:
            self.tool(tool)

    @property
    def persistent(self):
        return self._persistent

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def tools(self) -> list[Tool]:
        return list(self._tools)

    @property
    def skills(self) -> list[Skill]:
        return list(self._skills)

    @property
    def types(self):
        return (str, str)

    @property
    def llm(self):
        return self._llm

    async def perform(self, input: Message) -> AsyncIterator[Message]:
        from .context import Context
        """Main entrypoint for the agent.

        This method will select the right skill to perform the task and then execute it.
        The skill is selected based on the messages and the skills available to the agent.
        """
        context = Context(self, list(self._conversation) + [input])
        skill = await context.engage()

        messages = []

        async for m in skill.execute(context): # type: ignore
            yield m
            messages.append(m)

        self._conversation.extend(messages)

    def skill(self, target):
        """
        Add a method as a skill to the agent.
        The method must be an async generator.
        """
        if isinstance(target, Skill):
            self._skills.append(target)
            return target

        if not callable(target):
            raise ValueError("Skill must be a callable.")

        if not inspect.isasyncgenfunction(target):
            raise ValueError("Skill must be an async generator.")

        name = target.__name__
        description = inspect.getdoc(target) or ""
        skill = MethodSkill(name, description, target)
        self._skills.append(skill)
        return skill

    def tool(self, target):
        """
        Adds a method as a tool to the agent.

        If the method expects an LLM at any keyword parameter, the
        agent will automatically inject it.

        The method must be an async function.
        """

        if isinstance(target, Tool):
            self._tools.append(target)
            return target

        if not callable(target):
            raise ValueError("Tool must be a callable.")

        if not inspect.iscoroutinefunction(target):
            raise ValueError("Tool must be a coroutine function.")

        name = target.__name__
        description = inspect.getdoc(target) or ""
        signature = inspect.signature(target).parameters

        # If the method expects an LLM, wrap it
        # to inject the agent's LLM instance
        if any(issubclass(param.annotation, LLM) for param in signature.values()):
            target = self.llm.wrap(target)

        tool = MethodTool(name, description, target)
        self._tools.append(tool)
        return tool



from argo import ChatAgent, LLM, Context


SYSTEM_PROMPT = """
You are Lovelaice.

You are a smart, helpful AI agent that runs in the user terminal.
You can run bash commands, execute Python code, generate scripts, and
scaffold small projects.
You can also search the internet for questions of many kinds,
from trivia to programming to system engieering.
You can also answer questions about the user's files and code,
and manipulate the user filesytem, always with user confirmation.
"""


class Lovelaice(ChatAgent):
    def __init__(self, llm: LLM):
        super().__init__(
            "Lovelaice",
            "A helpful AI agent that runs in the user terminal.",
            llm,
            system_prompt=SYSTEM_PROMPT)

        # register all skills
        self.skill(self.chat)

        # register all tools

    async def chat(self, ctx: Context):
        """
        Chat with Lovelaice. Use this skill for casual chat.
        """
        yield await ctx.reply()



from fastapi import FastAPI
from fastapi.responses import StreamingResponse
from pydantic import BaseModel, create_model

from argo.tools import Tool

from .agent import ChatAgent
from .llm import Message


class SkillDescription(BaseModel):
    name: str
    description: str


class ToolDescription(BaseModel):
    name: str
    description: str
    parameters: dict[str, str]


class AgentDescription(BaseModel):
    name: str
    description: str
    skills: list[SkillDescription]
    tools: list[ToolDescription]


def build(agent: ChatAgent) -> FastAPI:
    """
    Builds a FastAPI app from an agent.

    This method sets up the following default routes:
     - `/` to return the agent's description.
     - `/chat` to perform the chat with the agent.

    It also sets up endpoints for each tool.

    The agent is stored in the app's state, so it can be accessed from the routes.
    """
    app = FastAPI()
    app.state.agent = agent

    @app.get("/")
    def info() -> AgentDescription:
        """
        Get the basic description of the agent.
        """
        return AgentDescription(
            name=agent.name,
            description=agent.description,
            skills=[SkillDescription(
                name=skill.name,
                description=skill.description,
            ) for skill in agent.skills],
            tools=[ToolDescription(
                name=tool.name,
                description=tool.description,
                parameters={ k:str(v) for k,v in tool.parameters() },
            ) for tool in agent.tools],
        )

    @app.post("/chat")
    async def chat(message: Message) -> Message:
        return await agent.perform(message)

    for tool in agent.tools:
        model_cls = build_model(tool)

        @app.post(f"/{tool.name}", response_model=model_cls)
        async def invoke_tool(params: model_cls) -> dict:
            return await tool.run(**params.dict())

    return app


def build_model(tool: Tool) -> type[BaseModel]:
    """
    Builds a Pydantic model from a tool.
    """
    return create_model(
        tool.name,
        **{k: v for k, v in tool.parameters()},
    )


def serve(agent: ChatAgent, host:str="127.0.0.1", port:int=8000):
    app = build(agent)
    import uvicorn
    uvicorn.run(app, host=host, port=port)
